@@ -1,18 +1,22 @@
 import logging
 import os
+import shlex
 import sys
 import traceback
 from argparse import Action, ArgumentParser, ArgumentTypeError, Namespace
 from itertools import chain
-from typing import Iterator, List, Optional, Tuple, Union
+from typing import Iterable, Iterator, List, Mapping, Optional, Tuple, Union
 
 import shtab
 
 from dql import __version__, utils
 from dql.catalog import Catalog
+from dql.cli_utils import BooleanOptionalAction
 from dql.client import Client
 from dql.data_storage import SQLiteDataStorage
 from dql.node import long_line_str
+from dql.remote.studio import StudioClient
+from dql.utils import DQLDir, get_remote_config, read_config
 
 logger = logging.getLogger("dql")
 
@@ -21,9 +25,7 @@ TTL_INT = 4 * 60 * 60
 FIND_COLUMNS = ["du", "name", "owner", "path", "size", "type"]
 
 
-def human_time_type(
-    value_str: str, can_be_none: bool = False
-) -> Optional[int]:
+def human_time_type(value_str: str, can_be_none: bool = False) -> Optional[int]:
     value = utils.human_time_to_int(value_str)
 
     if value:
@@ -32,8 +34,7 @@ def human_time_type(
         return None
 
     raise ArgumentTypeError(
-        "This option supports only a human-readable time "
-        "interval like 12h or 4w."
+        "This option supports only a human-readable time interval like 12h or 4w."
     )
 
 
@@ -42,8 +43,7 @@ def parse_find_column(column: str) -> str:
     if column_lower in FIND_COLUMNS:
         return column_lower
     raise ArgumentTypeError(
-        f"Invalid column for find: '{column}' "
-        f"Options are: {','.join(FIND_COLUMNS)}"
+        f"Invalid column for find: '{column}' Options are: {','.join(FIND_COLUMNS)}"
     )
 
 
@@ -57,9 +57,7 @@ def find_columns_type(
     return [parse_find_column(c) for c in columns_str.split(",")]
 
 
-def add_sources_arg(
-    parser: ArgumentParser, nargs: Union[str, int] = "+"
-) -> Action:
+def add_sources_arg(parser: ArgumentParser, nargs: Union[str, int] = "+") -> Action:
     return parser.add_argument(
         "sources",
         type=str,
@@ -71,9 +69,7 @@ def add_sources_arg(
 def get_parser() -> ArgumentParser:
     parser = ArgumentParser(description="DQL: Data Query Language", prog="dql")
 
-    parser.add_argument(
-        "-V", "--version", action="version", version=__version__
-    )
+    parser.add_argument("-V", "--version", action="version", version=__version__)
 
     parent_parser = ArgumentParser(add_help=False)
     parent_parser.add_argument(
@@ -103,9 +99,7 @@ def get_parser() -> ArgumentParser:
     )
 
     subp = parser.add_subparsers(help="Sub-command help", dest="command")
-    parse_get = subp.add_parser(
-        "get", parents=[parent_parser], help="Fetch a dataset"
-    )
+    parse_get = subp.add_parser("get", parents=[parent_parser], help="Fetch a dataset")
     parse_get.add_argument(
         "source", type=str, help="Data source - a path to a cloud storage dir"
     )
@@ -141,32 +135,111 @@ def get_parser() -> ArgumentParser:
         action="store_true",
         help="Copy directories recursively",
     )
-    parse_cp.add_argument(
-        "--dql-file",
-        help="Use a different filename for the resulting DQL file",
-    )
-    parse_cp.add_argument(
-        "--dql-only",
-        default=False,
-        action="store_true",
-        help=(
-            "Only create the resulting DQL file, "
-            "do not download or copy files"
-        ),
-    )
     parse_cp.add_argument(
         "--no-glob",
         default=False,
         action="store_true",
         help="Do not expand globs (such as * or ?)",
     )
-    parse_cp.add_argument(
+
+    parse_clone = subp.add_parser(
+        "clone", parents=[parent_parser], help="Copy data files from the cloud"
+    )
+    add_sources_arg(parse_clone).complete = shtab.DIR  # type: ignore[attr-defined] # noqa: E501
+    parse_clone.add_argument("output", type=str, help="Output")
+    parse_clone.add_argument(
+        "-f",
+        "--force",
+        default=False,
+        action="store_true",
+        help="Force creating outputs",
+    )
+    parse_clone.add_argument(
+        "-r",
+        "-R",
+        "--recursive",
+        default=False,
+        action="store_true",
+        help="Copy directories recursively",
+    )
+    parse_clone.add_argument(
+        "--dql-file",
+        help="Use a different filename for the resulting DQL file",
+    )
+    parse_clone.add_argument(
+        "--dql-only",
+        default=False,
+        action="store_true",
+        help="Only create the resulting DQL file, do not download or copy files",
+    )
+    parse_clone.add_argument(
+        "--no-glob",
+        default=False,
+        action="store_true",
+        help="Do not expand globs (such as * or ?)",
+    )
+    parse_clone.add_argument(
         "--no-dql-file",
         default=False,
         action="store_true",
         help="Do not write a .dql file",
     )
 
+    parse_register = subp.add_parser(
+        "register", parents=[parent_parser], help="Register a dataset"
+    )
+    parse_register.add_argument("name", type=str, help="Dataset name")
+    parse_register.add_argument(
+        "--description",
+        action="store",
+        default="",
+        help="Dataset description",
+    )
+    parse_register.add_argument(
+        "--labels",
+        default=[],
+        nargs="+",
+        help="Dataset labels",
+    )
+
+    parse_edit_dataset = subp.add_parser(
+        "edit-dataset", parents=[parent_parser], help="Edit dataset metadata"
+    )
+    parse_edit_dataset.add_argument("name", type=str, help="Dataset name")
+    parse_edit_dataset.add_argument(
+        "--new-name",
+        action="store",
+        default="",
+        help="Dataset new name",
+    )
+    parse_edit_dataset.add_argument(
+        "--description",
+        action="store",
+        default="",
+        help="Dataset description",
+    )
+    parse_edit_dataset.add_argument(
+        "--labels",
+        default=[],
+        nargs="+",
+        help="Dataset labels",
+    )
+
+    ls_datasets_parser = subp.add_parser(
+        "ls-datasets", parents=[parent_parser], help="List datasets"
+    )
+    ls_datasets_parser.add_argument(
+        "--shadow",
+        default=None,
+        action=BooleanOptionalAction,
+        help="List only shadow datasets, or only registered datasets",
+    )
+
+    rm_dataset_parser = subp.add_parser(
+        "rm-dataset", parents=[parent_parser], help="Removes dataset"
+    )
+    rm_dataset_parser.add_argument("name", type=str, help="Dataset name")
+
     parse_ls = subp.add_parser(
         "ls", parents=[parent_parser], help="List storage contents"
     )
@@ -178,6 +251,12 @@ def get_parser() -> ArgumentParser:
         default=0,
         help="List files in the long format",
     )
+    parse_ls.add_argument(
+        "--remote",
+        action="store",
+        default="",
+        help="Name of remote to use",
+    )
 
     parse_du = subp.add_parser(
         "du", parents=[parent_parser], help="Display space usage"
@@ -318,6 +397,19 @@ def cp(sources, output, **kwargs):
     catalog.cp(sources, output, **kwargs)
 
 
+def register(name, **kwargs):
+    data_storage = SQLiteDataStorage()
+    catalog = Catalog(data_storage)
+    catalog.register(name, **kwargs)
+    print(f"Registered dataset {name}, version 1 created")
+
+
+def edit_dataset(name, **kwargs):
+    data_storage = SQLiteDataStorage()
+    catalog = Catalog(data_storage)
+    catalog.edit_dataset(name, **kwargs)
+
+
 def ls_urls(
     sources,
     long: bool = False,
@@ -358,13 +450,12 @@ def _ls_urls_flat(
         if client_cls.is_root_url(source):
             buckets = client_cls.ls_buckets(**client_config)
             if long:
-                values = (
-                    long_line_str(b.name, b.created, "") for b in buckets
-                )
+                values = (long_line_str(b.name, b.created, "") for b in buckets)
             else:
                 values = (b.name for b in buckets)
             yield source, values
         else:
+            found = False
             for data_source, nodes in catalog.ls(
                 [source], client_config=client_config, **kwargs
             ):
@@ -372,7 +463,10 @@ def _ls_urls_flat(
                     values = (n.long_line_str() for n in nodes)
                 else:
                     values = (n.name_with_dir_ending for n in nodes)
-                yield "/".join(data_source.node.path), values
+                found = True
+                yield data_source.node.dirname, values
+            if not found:
+                raise FileNotFoundError(f"No such file or directory: {source}")
 
 
 def ls_indexed_storages(
@@ -390,18 +484,14 @@ def ls_indexed_storages(
             yield s.uri
 
 
-def ls(
-    sources, long: bool = False, catalog: Optional[Catalog] = None, **kwargs
-):
+def ls_local(sources, long: bool = False, catalog: Optional[Catalog] = None, **kwargs):
     if catalog is None:
         catalog = Catalog(SQLiteDataStorage())
     if sources:
         if len(sources) == 1:
-            for _, entries in ls_urls(
-                sources, long=long, catalog=catalog, **kwargs
-            ):
+            for _, entries in ls_urls(sources, long=long, catalog=catalog, **kwargs):
                 for entry in entries:
-                    print(entry)
+                    print(format_ls_entry(entry))
         else:
             first = True
             for source, entries in ls_urls(
@@ -415,10 +505,83 @@ def ls(
                 if source:
                     print(f"{source}:")
                 for entry in entries:
-                    print(entry)
+                    print(format_ls_entry(entry))
     else:
         for entry in ls_indexed_storages(long=long, catalog=catalog):
-            print(entry)
+            print(format_ls_entry(entry))
+
+
+def format_ls_entry(entry: str) -> str:
+    if entry.endswith("/") or not entry:
+        entry = shlex.quote(entry[:-1])
+        return f"{entry}/"
+    return shlex.quote(entry)
+
+
+def ls_remote(
+    url: str,
+    username: str,
+    token: str,
+    paths: Iterable[str],
+    long: bool = False,
+):
+    client = StudioClient(url, username, token)
+    first = True
+    for path, response in client.ls(paths):
+        if not first:
+            print()
+        if not response.ok or response.data is None:
+            print(f"{path}:\n  Error: {response.message}\n")
+            continue
+
+        print(f"{path}:")
+        if long:
+            for row in response.data:
+                entry = long_line_str(
+                    row["name"] + ("/" if row["dir_type"] else ""),
+                    row["last_modified"],
+                    row["owner_name"],
+                )
+                print(format_ls_entry(entry))
+        else:
+            for row in response.data:
+                entry = row["name"] + ("/" if row["dir_type"] else "")
+                print(format_ls_entry(entry))
+        first = False
+
+
+def ls(
+    sources,
+    long: bool = False,
+    remote: str = "",
+    config: Optional[Mapping[str, str]] = None,
+    **kwargs,
+):
+    if config is None:
+        config = get_remote_config(read_config(DQLDir.find().root), remote=remote)
+    remote_type = config["type"]
+    if remote_type == "local":
+        ls_local(sources, long=long, **kwargs)
+    else:
+        ls_remote(
+            config["url"],
+            config["username"],
+            config["token"],
+            sources,
+            long=long,
+        )
+
+
+def ls_datasets(shadow_only=None):
+    for d in Catalog(SQLiteDataStorage()).ls_datasets(shadow_only=shadow_only):
+        if d.shadow:
+            print(f"(tmp) {d.name}")
+        else:
+            print(d.name)
+
+
+def rm_dataset(name: str):
+    Catalog(SQLiteDataStorage()).remove_dataset(name)
 
 
 def du(sources, show_bytes=False, si=False, **kwargs):
@@ -479,6 +642,20 @@ def main(argv: Optional[List[str]] = None) -> int:  # noqa: C901
                 client_config=client_config,
             )
         elif args.command == "cp":
+            cp(
+                args.sources,
+                args.output,
+                force=bool(args.force),
+                update=bool(args.update),
+                recursive=bool(args.recursive),
+                dql_file=None,
+                dql_only=False,
+                no_glob=args.no_glob,
+                no_dql_file=True,
+                ttl=args.ttl,
+                client_config=client_config,
+            )
+        elif args.command == "clone":
             cp(
                 args.sources,
                 args.output,
@@ -488,17 +665,37 @@ def main(argv: Optional[List[str]] = None) -> int:  # noqa: C901
                 dql_file=args.dql_file,
                 dql_only=args.dql_only,
                 no_glob=args.no_glob,
+                no_dql_file=args.no_dql_file,
                 ttl=args.ttl,
                 client_config=client_config,
+                create_dataset=True,
+            )
+        elif args.command == "register":
+            register(
+                args.name,
+                description=args.description,
+                labels=args.labels,
+            )
+        elif args.command == "edit-dataset":
+            edit_dataset(
+                args.name,
+                description=args.description,
+                new_name=args.new_name,
+                labels=args.labels,
             )
         elif args.command == "ls":
             ls(
                 args.sources,
+                long=bool(args.long),
+                remote=args.remote,
                 ttl=args.ttl,
                 update=bool(args.update),
-                long=bool(args.long),
                 client_config=client_config,
             )
+        elif args.command == "ls-datasets":
+            ls_datasets(args.shadow)
+        elif args.command == "rm-dataset":
+            rm_dataset(args.name)
         elif args.command == "du":
             du(
                 args.sources,
