@@ -5,10 +5,7 @@ from functools import cached_property
 from typing import NamedTuple, Optional, Union
 from urllib.parse import urlparse
 
-from dateutil import tz
-from dateutil.parser import isoparse
-
-from dql.utils import time_to_str
+from dql.utils import is_expired, time_to_local_str, time_to_str
 
 
 class Status:
@@ -16,6 +13,7 @@ class Status:
     PENDING = 2
     FAILED = 3
     COMPLETE = 4
+    PARTIAL = 5
 
 
 class AbstractStorage(ABC):
@@ -62,11 +60,7 @@ class StorageRecord(NamedTuple):
 class Storage(StorageRecord, AbstractStorage):
     @property
     def is_expired(self) -> bool:
-        now = datetime.now()
-        if self.expires:
-            return self.time_to_local(self.expires) < self.time_to_local(now)
-
-        return False
+        return is_expired(self.expires)
 
     @property
     def timestamp_str(self) -> Optional[str]:
@@ -78,27 +72,13 @@ class Storage(StorageRecord, AbstractStorage):
     def timestamp_to_local(self) -> Optional[str]:
         if not self.timestamp:
             return None
-        return self.time_to_local_str(self.timestamp)
+        return time_to_local_str(self.timestamp)
 
     @property
     def expires_to_local(self) -> Optional[str]:
         if not self.expires:
             return None
-        return self.time_to_local_str(self.expires)
-
-    @staticmethod
-    def time_to_local_str(dt: Union[datetime, str]) -> str:
-        return time_to_str(Storage.time_to_local(dt))
-
-    @staticmethod
-    def time_to_local(dt: Union[datetime, str]) -> datetime:
-        # TODO check usage
-        if isinstance(dt, str):
-            dt = isoparse(dt)
-        try:
-            return dt.astimezone(tz.tzlocal())
-        except (OverflowError, OSError, ValueError):
-            return dt
+        return time_to_local_str(self.expires)
 
     @staticmethod
     def get_expiration_time(timestamp: datetime, ttl: int):
@@ -116,8 +96,6 @@ class Storage(StorageRecord, AbstractStorage):
             uri = posixpath.join(uri, *file_path.rstrip("/").split("/"))
         return {
             "uri": uri,
-            "timestamp": time_to_str(self.timestamp)
-            if self.timestamp
-            else None,
+            "timestamp": time_to_str(self.timestamp) if self.timestamp else None,
             "expires": time_to_str(self.expires) if self.expires else None,
         }
