@@ -10,10 +10,10 @@ from dql.catalog import Catalog
 from dql.data_storage import SQLiteDataStorage
 from dql.utils import DQLDir
 
+# pylint: disable=redefined-outer-name
+
 DEFAULT_DQL_BIN = "dql"
-DEFAULT_DQL_GIT_REPO = os.path.dirname(
-    os.path.dirname(os.path.abspath(__file__))
-)
+DEFAULT_DQL_GIT_REPO = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
 
 # JSON tests are currently ignored as they are a work in progress for SQLite
 collect_ignore = ["setup.py", "unit/test_json.py"]
@@ -52,7 +52,10 @@ def clean_environment(
 
 @pytest.fixture
 def data_storage():
-    _data_storage = SQLiteDataStorage(db_file=":memory:")
+    uri = "s3://somebucket"
+    _data_storage = SQLiteDataStorage(db_file=":memory:").clone(uri)
+    _data_storage.uri = uri
+    _data_storage.init_db()
     yield _data_storage
 
     # Check that the DB isn't locked after the test
@@ -79,10 +82,7 @@ def pytest_addoption(parser):
     parser.addoption(
         "--dql-revs",
         type=str,
-        help=(
-            "Comma-separated list of DQL revisions to test "
-            "(overrides `--dql-bin`)"
-        ),
+        help="Comma-separated list of DQL revisions to test (overrides `--dql-bin`)",
     )
 
     parser.addoption(
@@ -91,6 +91,14 @@ def pytest_addoption(parser):
         default=DEFAULT_DQL_GIT_REPO,
         help="Path or url to dql git repo",
     )
+    parser.addoption(
+        "--azure-connection-string",
+        type=str,
+        help=(
+            "Connection string to run tests against a real, versioned "
+            "Azure storage account"
+        ),
+    )
 
 
 class DQLTestConfig:
@@ -113,6 +121,34 @@ def pytest_configure(config):
     config.dql_config.dql_git_repo = config.getoption("--dql-git-repo")
 
 
+class CloudTestServer(NamedTuple):
+    src: CloudPath
+    client_config: Dict[str, str]
+
+
+def make_cloud_server(src_path, cloud_type):
+    fs = src_path.fs
+    if cloud_type == "s3":
+        endpoint_url = fs.client_kwargs["endpoint_url"]
+        client_config = {"aws_endpoint_url": endpoint_url}
+    elif cloud_type == "gcs":
+        endpoint_url = fs._endpoint  # pylint:disable=protected-access
+        client_config = {"endpoint_url": endpoint_url}
+    elif cloud_type == "azure":
+        client_config = fs.storage_options.copy()
+    else:
+        raise ValueError(f"invalid cloud_type: {cloud_type}")
+    (src_path / "description").write_text("Cats and Dogs")
+    (src_path / "cats" / "cat1").write_text("meow")
+    (src_path / "cats" / "cat2").write_text("mrow")
+    (src_path / "dogs" / "dog1").write_text("woof")
+    (src_path / "dogs" / "dog2").write_text("arf")
+    (src_path / "dogs" / "dog3").write_text("bark")
+    (src_path / "dogs" / "others" / "dog4").write_text("ruff")
+
+    return CloudTestServer(src=src_path, client_config=client_config)
+
+
 class CloudTestCatalog(NamedTuple):
     src: CloudPath
     working_dir: PosixPath
@@ -120,61 +156,45 @@ class CloudTestCatalog(NamedTuple):
     client_config: Dict[str, str]
 
 
-class CloudTestServer(NamedTuple):
-    src: CloudPath
-    client_config: Dict[str, str]
+@pytest.fixture(scope="session", params=["s3", "gcs", "azure"])
+def cloud_type(request):
+    return request.param
 
 
-# Test servers can be cached as they are not modified in these tests
-cloud_server_cache: Dict[str, CloudTestServer] = {}
+@pytest.fixture(scope="session", params=[False, True])
+def version_aware(request):
+    return request.param
+
+
+@pytest.fixture(scope="session")
+def cloud_server(request, tmp_upath_factory, cloud_type, version_aware):
+    if cloud_type == "azure" and version_aware:
+        if conn_str := request.config.getoption("--azure-connection-string"):
+            src_path = tmp_upath_factory.azure(conn_str)
+        else:
+            pytest.skip("Can't test versioning with Azure")
+    else:
+        src_path = tmp_upath_factory.mktemp(cloud_type, version_aware=version_aware)
+    return make_cloud_server(src_path, cloud_type)
 
 
 @pytest.fixture
 def cloud_test_catalog(
-    request,
+    cloud_server,
     tmp_path,
-    tmp_upath_factory,
     data_storage,  # pylint: disable=redefined-outer-name # noqa: E501
 ):
-    cloud_type = request.param
-    if cloud_type in cloud_server_cache:
-        src_path = cloud_server_cache[cloud_type].src
-        client_config = cloud_server_cache[cloud_type].client_config
-    else:
-        src_path = tmp_upath_factory.mktemp(cloud_type)
-        fs = src_path._accessor._fs  # pylint:disable=protected-access
-        if cloud_type == "s3":
-            endpoint_url = fs.client_kwargs["endpoint_url"]
-            client_config = {"aws_endpoint_url": endpoint_url}
-        elif cloud_type == "gcs":
-            endpoint_url = fs._endpoint  # pylint:disable=protected-access
-            client_config = {"endpoint_url": endpoint_url}
-        else:
-            raise ValueError(f"invalid cloud_type: {cloud_type}")
-        (src_path / "description").write_text("Cats and Dogs")
-        (src_path / "cats" / "cat1").write_text("meow")
-        (src_path / "cats" / "cat2").write_text("mrow")
-        (src_path / "dogs" / "dog1").write_text("woof")
-        (src_path / "dogs" / "dog2").write_text("arf")
-        (src_path / "dogs" / "dog3").write_text("bark")
-        (src_path / "dogs" / "others" / "dog4").write_text("ruff")
-        cloud_server_cache[cloud_type] = CloudTestServer(
-            src=src_path,
-            client_config=client_config,
-        )
-
     cache_dir = tmp_path / ".dql" / "cache"
     cache_dir.mkdir(parents=True)
     tmpfile_dir = tmp_path / ".dql" / "tmp"
     tmpfile_dir.mkdir()
 
-    catalog = Catalog(
-        data_storage, cache_dir=str(cache_dir), tmp_dir=str(tmpfile_dir)
-    )
+    data_storage = data_storage.clone(str(cloud_server.src).rstrip("/"))
+    catalog = Catalog(data_storage, cache_dir=str(cache_dir), tmp_dir=str(tmpfile_dir))
 
     return CloudTestCatalog(
-        src=src_path,
+        src=cloud_server.src,
         working_dir=tmp_path,
         catalog=catalog,
-        client_config=client_config,
+        client_config=cloud_server.client_config,
     )
