@@ -1,4 +1,6 @@
 import shutil
+import sqlite3
+import uuid
 
 import pytest
 import yaml
@@ -7,7 +9,6 @@ from dql.catalog import parse_dql_file
 from dql.utils import remove_readonly
 
 
-@pytest.mark.parametrize("cloud_test_catalog", ["s3", "gcs"], indirect=True)
 def test_find(cloud_test_catalog):
     src = cloud_test_catalog.src
     catalog = cloud_test_catalog.catalog
@@ -35,40 +36,28 @@ def test_find(cloud_test_catalog):
         )
 
 
-@pytest.mark.parametrize("cloud_test_catalog", ["s3", "gcs"], indirect=True)
 def test_find_names_paths_size_type(cloud_test_catalog):
     src = cloud_test_catalog.src
     catalog = cloud_test_catalog.catalog
     config = cloud_test_catalog.client_config
 
-    assert set(
-        catalog.find([str(src)], names=["*cat*"], client_config=config)
-    ) == {
+    assert set(catalog.find([str(src)], names=["*cat*"], client_config=config)) == {
         str(src / "cats") + "/",
         str(src / "cats" / "cat1"),
         str(src / "cats" / "cat2"),
     }
 
     assert set(
-        catalog.find(
-            [str(src)], names=["*cat*"], typ="dir", client_config=config
-        )
+        catalog.find([str(src)], names=["*cat*"], typ="dir", client_config=config)
     ) == {
         str(src / "cats") + "/",
     }
 
     assert (
-        len(
-            list(
-                catalog.find([str(src)], names=["*CAT*"], client_config=config)
-            )
-        )
-        == 0
+        len(list(catalog.find([str(src)], names=["*CAT*"], client_config=config))) == 0
     )
 
-    assert set(
-        catalog.find([str(src)], inames=["*CAT*"], client_config=config)
-    ) == {
+    assert set(catalog.find([str(src)], inames=["*CAT*"], client_config=config)) == {
         str(src / "cats") + "/",
         str(src / "cats" / "cat1"),
         str(src / "cats" / "cat2"),
@@ -82,13 +71,7 @@ def test_find_names_paths_size_type(cloud_test_catalog):
     }
 
     assert (
-        len(
-            list(
-                catalog.find(
-                    [str(src)], paths=["*caTS/CaT**"], client_config=config
-                )
-            )
-        )
+        len(list(catalog.find([str(src)], paths=["*caTS/CaT**"], client_config=config)))
         == 0
     )
 
@@ -99,20 +82,15 @@ def test_find_names_paths_size_type(cloud_test_catalog):
         str(src / "cats" / "cat2"),
     }
 
-    assert set(
-        catalog.find([str(src)], size="5", typ="f", client_config=config)
-    ) == {
+    assert set(catalog.find([str(src)], size="5", typ="f", client_config=config)) == {
         str(src / "description"),
     }
 
-    assert set(
-        catalog.find([str(src)], size="-3", typ="f", client_config=config)
-    ) == {
+    assert set(catalog.find([str(src)], size="-3", typ="f", client_config=config)) == {
         str(src / "dogs" / "dog2"),
     }
 
 
-@pytest.mark.parametrize("cloud_test_catalog", ["s3", "gcs"], indirect=True)
 def test_find_names_columns(cloud_test_catalog):
     src = cloud_test_catalog.src
     catalog = cloud_test_catalog.catalog
@@ -138,12 +116,17 @@ def test_find_names_columns(cloud_test_catalog):
     }
 
 
-@pytest.mark.parametrize("cloud_test_catalog", ["s3", "gcs"], indirect=True)
 @pytest.mark.parametrize(
-    "recursive,star",
-    ((True, True), (True, False), (False, True), (False, False)),
+    "recursive,star,dir_exists",
+    (
+        (True, True, False),
+        (True, False, False),
+        (True, False, True),
+        (False, True, False),
+        (False, False, False),
+    ),
 )
-def test_cp_root(cloud_test_catalog, recursive, star):
+def test_cp_root(cloud_test_catalog, recursive, star, dir_exists):
     src = cloud_test_catalog.src
     working_dir = cloud_test_catalog.working_dir
     catalog = cloud_test_catalog.catalog
@@ -155,6 +138,18 @@ def test_cp_root(cloud_test_catalog, recursive, star):
     else:
         src_path = str(src)
 
+    if star:
+        with pytest.raises(FileNotFoundError):
+            catalog.cp(
+                [src_path],
+                str(dest),
+                client_config=cloud_test_catalog.client_config,
+                recursive=recursive,
+            )
+
+    if dir_exists or star:
+        dest.mkdir()
+
     catalog.cp(
         [src_path],
         str(dest),
@@ -194,7 +189,7 @@ def test_cp_root(cloud_test_catalog, recursive, star):
         assert len(f["checksum"]) > 1
 
     # Description is always copied (if anything is copied)
-    prefix = "" if star else "/"
+    prefix = "" if star or (recursive and not dir_exists) else "/"
     assert files_by_name[f"{prefix}description"]["size"] == 13
 
     if recursive:
@@ -223,19 +218,19 @@ def test_cp_root(cloud_test_catalog, recursive, star):
         assert f"{prefix}dogs/others/dog4" not in files_by_name
 
 
-@pytest.mark.parametrize("cloud_test_catalog", ["s3", "gcs"], indirect=True)
 @pytest.mark.parametrize(
-    "recursive,star,slash",
+    "recursive,star,slash,dir_exists",
     (
-        (True, True, False),
-        (True, False, False),
-        (True, False, True),
-        (False, True, False),
-        (False, False, False),
-        (False, False, True),
+        (True, True, False, False),
+        (True, False, False, False),
+        (True, False, False, True),
+        (True, False, True, False),
+        (False, True, False, False),
+        (False, False, False, False),
+        (False, False, True, False),
     ),
 )
-def test_cp_subdir(cloud_test_catalog, recursive, star, slash):
+def test_cp_subdir(cloud_test_catalog, recursive, star, slash, dir_exists):
     src = cloud_test_catalog.src / "dogs"
     working_dir = cloud_test_catalog.working_dir
     catalog = cloud_test_catalog.catalog
@@ -249,6 +244,18 @@ def test_cp_subdir(cloud_test_catalog, recursive, star, slash):
         if slash:
             src_path = f"{src_path}/"
 
+    if star:
+        with pytest.raises(FileNotFoundError):
+            catalog.cp(
+                [src_path],
+                str(dest),
+                client_config=cloud_test_catalog.client_config,
+                recursive=recursive,
+            )
+
+    if dir_exists or star:
+        dest.mkdir()
+
     catalog.cp(
         [src_path],
         str(dest),
@@ -284,7 +291,7 @@ def test_cp_subdir(cloud_test_catalog, recursive, star, slash):
     for f in data["files"]:
         assert len(f["checksum"]) > 1
 
-    if star or slash:
+    if not dir_exists:
         assert (dest / "dog1").read_text() == "woof"
         assert (dest / "dog2").read_text() == "arf"
         assert (dest / "dog3").read_text() == "bark"
@@ -314,7 +321,6 @@ def test_cp_subdir(cloud_test_catalog, recursive, star, slash):
     assert files_by_name["dogs/others/dog4"]["size"] == 4
 
 
-@pytest.mark.parametrize("cloud_test_catalog", ["s3", "gcs"], indirect=True)
 @pytest.mark.parametrize(
     "recursive,star,slash",
     (
@@ -343,6 +349,16 @@ def test_cp_multi_subdir(cloud_test_catalog, recursive, star, slash):
         if slash:
             src_paths = [f"{src}/" for src in src_paths]
 
+    with pytest.raises(FileNotFoundError):
+        catalog.cp(
+            src_paths,
+            str(dest),
+            client_config=cloud_test_catalog.client_config,
+            recursive=recursive,
+        )
+
+    dest.mkdir()
+
     catalog.cp(
         src_paths,
         str(dest),
@@ -427,7 +443,6 @@ def test_cp_multi_subdir(cloud_test_catalog, recursive, star, slash):
     assert dog_files_by_name["dogs/others/dog4"]["size"] == 4
 
 
-@pytest.mark.parametrize("cloud_test_catalog", ["s3", "gcs"], indirect=True)
 def test_cp_double_subdir(cloud_test_catalog):
     src = cloud_test_catalog.src / "dogs" / "others"
     working_dir = cloud_test_catalog.working_dir
@@ -467,7 +482,34 @@ def test_cp_double_subdir(cloud_test_catalog):
     assert files_by_name["dog4"]["size"] == 4
 
 
-@pytest.mark.parametrize("cloud_test_catalog", ["s3", "gcs"], indirect=True)
+@pytest.mark.parametrize("no_glob", (True, False))
+def test_cp_single_file(cloud_test_catalog, no_glob):
+    src = cloud_test_catalog.src / "dogs" / "dog1"
+    working_dir = cloud_test_catalog.working_dir
+    catalog = cloud_test_catalog.catalog
+
+    dest = working_dir / "data"
+
+    src_path = str(src)
+
+    print(src_path)
+
+    dest.mkdir()
+
+    catalog.cp(
+        [src_path],
+        str(dest / "local_dog"),
+        client_config=cloud_test_catalog.client_config,
+        no_dql_file=True,
+        no_glob=no_glob,
+    )
+
+    assert dest.exists()
+    assert (dest / "dogs").exists() is False
+    assert (dest / "others").exists() is False
+    assert (dest / "local_dog").read_text() == "woof"
+
+
 def test_cp_dql_file_options(cloud_test_catalog):
     src = cloud_test_catalog.src / "dogs"
     working_dir = cloud_test_catalog.working_dir
@@ -546,7 +588,6 @@ def test_cp_dql_file_options(cloud_test_catalog):
     assert dql_only_source["files"] == data["files"]
 
 
-@pytest.mark.parametrize("cloud_test_catalog", ["s3", "gcs"], indirect=True)
 def test_cp_dql_file_sources(cloud_test_catalog):
     sources = [
         f"{str(cloud_test_catalog.src / 'cats')}/",
@@ -619,6 +660,18 @@ def test_cp_dql_file_sources(cloud_test_catalog):
     assert dog_files_by_name1["dog3"]["size"] == 4
     assert dog_files_by_name1["others/dog4"]["size"] == 4
 
+    assert not dest.exists()
+
+    with pytest.raises(FileNotFoundError):
+        catalog.cp(
+            [str(dqf) for dqf in dql_files],
+            str(dest),
+            client_config=cloud_test_catalog.client_config,
+            recursive=True,
+        )
+
+    dest.mkdir()
+
     # Copy using these DQL files as sources
     catalog.cp(
         [str(dqf) for dqf in dql_files],
@@ -666,7 +719,6 @@ def test_cp_dql_file_sources(cloud_test_catalog):
     assert dog_files_by_name2["others/dog4"]["size"] == 4
 
 
-@pytest.mark.parametrize("cloud_test_catalog", ["s3", "gcs"], indirect=True)
 def test_get(cloud_test_catalog):
     src = cloud_test_catalog.src
     catalog = cloud_test_catalog.catalog
@@ -684,7 +736,6 @@ def test_get(cloud_test_catalog):
     assert dest.with_suffix(".dql").is_file()
 
 
-@pytest.mark.parametrize("cloud_test_catalog", ["s3", "gcs"], indirect=True)
 def test_get_subdir(cloud_test_catalog):
     src = f"{str(cloud_test_catalog.src)}/dogs"
     working_dir = cloud_test_catalog.working_dir
@@ -707,21 +758,16 @@ def test_get_subdir(cloud_test_catalog):
 
     with pytest.raises(RuntimeError):
         # An error should be raised if the output directory already exists
-        catalog.get(
-            src, str(dest), client_config=cloud_test_catalog.client_config
-        )
+        catalog.get(src, str(dest), client_config=cloud_test_catalog.client_config)
 
     shutil.rmtree(dest, onerror=remove_readonly)
     assert dest.with_suffix(".dql").is_file()
 
     with pytest.raises(RuntimeError):
         # An error should also be raised if the dataset file already exists
-        catalog.get(
-            src, str(dest), client_config=cloud_test_catalog.client_config
-        )
+        catalog.get(src, str(dest), client_config=cloud_test_catalog.client_config)
 
 
-@pytest.mark.parametrize("cloud_test_catalog", ["s3", "gcs"], indirect=True)
 def test_du(cloud_test_catalog):
     src = str(cloud_test_catalog.src)
     catalog = cloud_test_catalog.catalog
@@ -737,20 +783,15 @@ def test_du(cloud_test_catalog):
 
     assert list(results) == expected_results[3:]
 
-    results = catalog.du(
-        [src], client_config=cloud_test_catalog.client_config, depth=1
-    )
+    results = catalog.du([src], client_config=cloud_test_catalog.client_config, depth=1)
 
     assert list(results) == expected_results[:1] + expected_results[2:]
 
-    results = catalog.du(
-        [src], client_config=cloud_test_catalog.client_config, depth=5
-    )
+    results = catalog.du([src], client_config=cloud_test_catalog.client_config, depth=5)
 
     assert list(results) == expected_results
 
 
-@pytest.mark.parametrize("cloud_test_catalog", ["s3", "gcs"], indirect=True)
 def test_ls_glob(cloud_test_catalog):
     src = cloud_test_catalog.src
     catalog = cloud_test_catalog.catalog
@@ -762,3 +803,223 @@ def test_ls_glob(cloud_test_catalog):
             client_config=cloud_test_catalog.client_config,
         )
     ] == [("dog1", ["dog1"]), ("dog2", ["dog2"]), ("dog3", ["dog3"])]
+
+
+def test_upserting_shadow_dataset(cloud_test_catalog):
+    shadow_dataset_name = uuid.uuid4().hex
+    src = cloud_test_catalog.src
+    catalog = cloud_test_catalog.catalog
+
+    # list source to have it in db as source for dataset
+    list(
+        catalog.ls(
+            [str(src)],
+            client_config=cloud_test_catalog.client_config,
+        )
+    )
+
+    catalog.upsert_shadow_dataset(
+        shadow_dataset_name,
+        [str(src / "dogs" / "*")],
+        client_config=cloud_test_catalog.client_config,
+        recursive=True,
+    )
+
+    dataset = catalog.data_storage.get_dataset(shadow_dataset_name)
+    assert dataset.name == shadow_dataset_name
+    assert dataset.description is None
+    assert dataset.version is None
+    assert dataset.labels == []
+    assert dataset.shadow is True
+
+    dataset_table_name = catalog.data_storage._dataset_table_name(dataset.id)
+    data = catalog.data_storage.db.execute(
+        f"select * from {dataset_table_name}"
+    ).fetchall()
+    assert data
+
+
+def test_registering_dataset(cloud_test_catalog):
+    shadow_dataset_name = uuid.uuid4().hex
+    src = cloud_test_catalog.src
+    catalog = cloud_test_catalog.catalog
+
+    # list source to have it in db as source for dataset
+    list(
+        catalog.ls(
+            [str(src)],
+            client_config=cloud_test_catalog.client_config,
+        )
+    )
+
+    catalog.upsert_shadow_dataset(
+        shadow_dataset_name,
+        [str(src / "dogs" / "*")],
+        client_config=cloud_test_catalog.client_config,
+        recursive=True,
+    )
+
+    catalog.register(
+        shadow_dataset_name,
+        description="dogs dataset",
+        labels=["dogs", "dataset"],
+    )
+
+    dataset = catalog.data_storage.get_dataset(shadow_dataset_name)
+    dataset_table_name = catalog.data_storage._dataset_table_name(dataset.id)
+    assert dataset.name == shadow_dataset_name
+    assert dataset.description == "dogs dataset"
+    assert dataset.version == "1"
+    assert dataset.labels == ["dogs", "dataset"]
+    assert dataset.shadow is False
+    data = catalog.data_storage.db.execute(
+        f"select * from {dataset_table_name}"
+    ).fetchall()
+    assert data
+
+
+def test_registering_dataset_with_new_name(cloud_test_catalog):
+    shadow_dataset_name = uuid.uuid4().hex
+    new_dataset_name = uuid.uuid4().hex
+    src = cloud_test_catalog.src
+    catalog = cloud_test_catalog.catalog
+
+    # list source to have it in db as source for dataset
+    list(
+        catalog.ls(
+            [str(src)],
+            client_config=cloud_test_catalog.client_config,
+        )
+    )
+
+    catalog.upsert_shadow_dataset(
+        shadow_dataset_name,
+        [str(src / "dogs" / "*")],
+        client_config=cloud_test_catalog.client_config,
+        recursive=True,
+    )
+
+    catalog.register(
+        shadow_dataset_name,
+        new_dataset_name=new_dataset_name,
+        description="dogs dataset",
+        labels=["dogs", "dataset"],
+    )
+    dataset = catalog.data_storage.get_dataset(new_dataset_name)
+    assert dataset
+    dataset_table_name = catalog.data_storage._dataset_table_name(dataset.id)
+    assert dataset.name == new_dataset_name
+    data = catalog.data_storage.db.execute(
+        f"select * from {dataset_table_name}"
+    ).fetchall()
+    assert data
+
+
+def test_registering_dataset_with_custom_version(cloud_test_catalog):
+    shadow_dataset_name = uuid.uuid4().hex
+    src = cloud_test_catalog.src
+    catalog = cloud_test_catalog.catalog
+
+    # list source to have it in db as source for dataset
+    list(
+        catalog.ls(
+            [str(src)],
+            client_config=cloud_test_catalog.client_config,
+        )
+    )
+
+    catalog.upsert_shadow_dataset(
+        shadow_dataset_name,
+        [str(src / "dogs" / "*")],
+        client_config=cloud_test_catalog.client_config,
+        recursive=True,
+    )
+
+    catalog.register(
+        shadow_dataset_name,
+        version="0.0.1",
+        description="dogs dataset",
+        labels=["dogs", "dataset"],
+    )
+
+    dataset = catalog.data_storage.get_dataset(shadow_dataset_name)
+    assert dataset.version == "0.0.1"
+
+
+def test_removing_dataset(cloud_test_catalog):
+    shadow_dataset_name = uuid.uuid4().hex
+    src = cloud_test_catalog.src
+    catalog = cloud_test_catalog.catalog
+
+    # list source to have it in db as source for dataset
+    list(
+        catalog.ls(
+            [str(src)],
+            client_config=cloud_test_catalog.client_config,
+        )
+    )
+
+    catalog.upsert_shadow_dataset(
+        shadow_dataset_name,
+        [str(src / "dogs" / "*")],
+        client_config=cloud_test_catalog.client_config,
+        recursive=True,
+    )
+
+    dataset = catalog.data_storage.get_dataset(shadow_dataset_name)
+    assert dataset
+    dataset_table_name = catalog.data_storage._dataset_table_name(dataset.id)
+    data = catalog.data_storage.db.execute(
+        f"select * from {dataset_table_name}"
+    ).fetchall()
+    assert data
+
+    catalog.remove_dataset(shadow_dataset_name)
+    dataset = catalog.data_storage.get_dataset(shadow_dataset_name)
+    assert dataset is None
+    with pytest.raises(sqlite3.OperationalError):
+        catalog.data_storage.db.execute(
+            f"select * from {dataset_table_name}"
+        ).fetchall()
+
+
+def test_edit_dataset(cloud_test_catalog):
+    shadow_dataset_name = uuid.uuid4().hex
+    dataset_new_name = uuid.uuid4().hex
+    src = cloud_test_catalog.src
+    catalog = cloud_test_catalog.catalog
+
+    # list source to have it in db as source for dataset
+    list(
+        catalog.ls(
+            [str(src)],
+            client_config=cloud_test_catalog.client_config,
+        )
+    )
+
+    catalog.upsert_shadow_dataset(
+        shadow_dataset_name,
+        [str(src / "dogs" / "*")],
+        client_config=cloud_test_catalog.client_config,
+        recursive=True,
+    )
+
+    catalog.register(
+        shadow_dataset_name,
+        version="0.0.1",
+        description="dogs dataset",
+        labels=["dogs", "dataset"],
+    )
+
+    catalog.edit_dataset(
+        shadow_dataset_name,
+        new_name=dataset_new_name,
+        description="new description",
+        labels=["cats", "birds"],
+    )
+
+    dataset = catalog.data_storage.get_dataset(dataset_new_name)
+    assert dataset.version == "0.0.1"
+    assert dataset.name == dataset_new_name
+    assert dataset.description == "new description"
+    assert dataset.labels == ["cats", "birds"]
