@@ -26,9 +26,7 @@ def test_storage():
 
 
 def test_expiration_time():
-    assert Storage.get_expiration_time(TS, 12344) == TS + timedelta(
-        seconds=12344
-    )
+    assert Storage.get_expiration_time(TS, 12344) == TS + timedelta(seconds=12344)
 
 
 def test_adding_storage(data_storage):
@@ -39,7 +37,7 @@ def test_adding_storage(data_storage):
     cached_storage = data_storage.get_storage(uri)
     assert cached_storage is None
 
-    storage, _, _ = data_storage.register_storage_for_indexing(uri)
+    storage, _, _, _ = data_storage.register_storage_for_indexing(uri)
     cnt = data_storage.db.execute("SELECT COUNT() FROM buckets").fetchall()
     assert cnt[0][0] == 1
 
@@ -52,7 +50,7 @@ def test_adding_storage(data_storage):
 
 
 def test_storage_status(data_storage):
-    uri = "s3://somebucket"
+    uri = data_storage.uri
 
     data_storage.create_storage_if_not_registered(uri)
     storage = data_storage.get_storage(uri)
@@ -63,20 +61,23 @@ def test_storage_status(data_storage):
         storage,
         need_index,
         in_progress,
+        is_new,
     ) = data_storage.register_storage_for_indexing(uri)
     assert storage.status == Status.PENDING
     assert storage.uri == uri
     assert storage == data_storage.get_storage(uri)
     assert need_index is True
     assert in_progress is False
+    assert is_new is True
 
-    s2, need_index, in_progress = data_storage.register_storage_for_indexing(
+    s2, need_index, in_progress, is_new = data_storage.register_storage_for_indexing(
         uri
     )
     assert s2.status == Status.PENDING
     assert storage == s2 == data_storage.get_storage(uri)
     assert need_index is False
     assert in_progress is True
+    assert is_new is False
 
     end_time = datetime.now(timezone.utc)
     data_storage.mark_storage_indexed(uri, Status.COMPLETE, 1000, end_time)
